/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */


#ifndef _NV_LINUX_H_
#define _NV_LINUX_H_

#include "nv.h"

#include <linux/config.h>
#include <linux/version.h>


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
#  error This driver does not support pre-2.4 kernels!
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#  define KERNEL_2_4
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#  error This driver does not support 2.5 kernels!
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 7, 0)
#  define KERNEL_2_6
#else
#  error This driver does not support development kernels!
#endif

#if defined (__ia64)
#  if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 13)
#    error This driver does not support 2.4.12 or earlier kernels!
#  endif
#endif

#if defined (CONFIG_SMP) && !defined (__SMP__)
#define __SMP__
#endif

#if defined (CONFIG_MODVERSIONS) && !defined (MODVERSIONS)
#  define MODVERSIONS
#endif

#if defined (MODVERSIONS) && !defined (KERNEL_2_6)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>

#include <linux/init.h>             /* module_init, module_exit         */
#include <linux/types.h>            /* pic_t, size_t, __u32, etc        */
#include <linux/errno.h>            /* error codes                      */
#include <linux/list.h>             /* circular linked list             */
#include <linux/stddef.h>           /* NULL, offsetof                   */
#include <linux/wait.h>             /* wait queues                      */

#include <linux/slab.h>             /* kmalloc, kfree, etc              */
#include <linux/vmalloc.h>          /* vmalloc, vfree, etc              */

#include <linux/poll.h>             /* poll_wait                        */
#include <linux/delay.h>            /* mdelay, udelay                   */

#ifdef KERNEL_2_6
#include <linux/sched.h>            /* suser(), capable() replacement   */
#include <linux/moduleparam.h>      /* module_param()                   */
#include <linux/smp_lock.h>         /* kernel_locked                    */
#include <asm/kmap_types.h>         /* page table entry lookup          */
#endif

#include <linux/pci.h>              /* pci_find_class, etc              */
#include <linux/interrupt.h>        /* tasklets, interrupt helpers      */
#include <linux/timer.h>

#include <asm/system.h>             /* cli, sli, save_flags             */
#include <asm/io.h>                 /* ioremap, virt_to_phys            */
#include <asm/uaccess.h>            /* access_ok                        */
#include <asm/page.h>               /* PAGE_OFFSET                      */
#include <asm/pgtable.h>            /* pte bit definitions              */

#include <linux/spinlock.h>
#include <asm/semaphore.h>
#include <linux/highmem.h>

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
#endif

#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif

#ifdef CONFIG_PM
#include <linux/pm.h>
#endif

#ifdef CONFIG_MTRR
#include <asm/mtrr.h>
#endif

#ifdef CONFIG_KDB
#include <linux/kdb.h>
#include <asm/kdb.h>
#endif

#if defined (CONFIG_AGP) || defined (CONFIG_AGP_MODULE)
#ifndef NOAGPGART
#  define AGPGART
#  include <linux/agp_backend.h>
#  include <linux/agpgart.h>
#endif
#endif

#if !defined (list_for_each)
#define list_for_each(pos, head) \
        for (pos = (head)->next; pos != (head); pos = (pos)->next)
#endif


#ifndef NVWATCH

/* various memory tracking/debugging techniques
 * disabled for retail builds, enabled for debug builds
 */

#ifdef DEBUG
#define NV_ENABLE_MEM_TRACKING 1
#endif

#if NV_ENABLE_MEM_TRACKING
#define NV_MEM_TRACKING_PAD_SIZE(size)   ((size) += sizeof(void *))
#define NV_MEM_TRACKING_HIDE_SIZE(ptr, size) \
    if ((ptr) && *(ptr)) { \
        *(int *) *(ptr) = (size); \
        (char *) *(ptr) += sizeof(void *); \
    }
#define NV_MEM_TRACKING_RETRIEVE_SIZE(ptr, size) \
    { \
        (char *) (ptr) -= sizeof(void *); \
        size = *(int *) (ptr); \
    }
#else
#define NV_MEM_TRACKING_PAD_SIZE(size)
#define NV_MEM_TRACKING_HIDE_SIZE(ptr, size)
#define NV_MEM_TRACKING_RETRIEVE_SIZE(ptr, size)  ((size) = 0)
#endif


/* poor man's memory allocation tracker.
 * main intention is just to see how much memory is being used to recognize
 * when memory usage gets out of control or if memory leaks are happening
 */

/* keep track of memory usage */
#if NV_ENABLE_MEM_TRACKING
extern int vm_usage;
extern int km_usage;
extern int fp_usage;

/* print out a running tally of memory allocation amounts, disabled by default */
// #define POOR_MANS_MEM_CHECK 1


/* slightly more advanced memory allocation tracker.
 * track who's allocating memory and print out a list of currently allocated
 * memory at key points in the driver
 */

#define MEMDBG_ALLOC(a,b) (a = kmalloc(b, GFP_ATOMIC))
#define MEMDBG_FREE(a)    (kfree(a))

#include "nv-memdbg.h"

#undef MEMDBG_ALLOC
#undef MEMDBG_FREE

extern struct mem_track_t *vm_list;
extern struct mem_track_t *km_list;
extern struct mem_track_t *fp_list;

/* print out list of memory allocations */
/* default to enabled for now */
#define LIST_MEM_CHECK 1

/* decide which memory types to apply mem trackers to */
#define VM_CHECKER 1
#define KM_CHECKER 1
#define FP_CHECKER 1

#endif  /* NV_ENABLE_MEM_TRACKING */

#define VM_PRINT(str, args...)
#define KM_PRINT(str, args...)
#define FP_PRINT(str, args...)

#define VM_ADD_MEM(a,b,c,d)
#define VM_FREE_MEM(a,b,c,d)
#define KM_ADD_MEM(a,b,c,d)
#define KM_FREE_MEM(a,b,c,d)
#define FP_ADD_MEM(a,b,c,d)
#define FP_FREE_MEM(a,b,c,d)

#define VM_ALLOC_RECORD(a,b,c)
#define VM_FREE_RECORD(a,b,c)
#define KM_ALLOC_RECORD(a,b,c)
#define KM_FREE_RECORD(a,b,c)


#if defined(VM_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  VM_PRINT
#    define VM_PRINT(str, args...)   printk(str, ##args)
#  endif
#  if defined(LIST_MEM_CHECK)
#    undef  VM_ADD_MEM
#    define VM_ADD_MEM(a,b,c,d)      nv_add_mem(&vm_list, a, b, c, d)
#    undef  VM_FREE_MEM
#    define VM_FREE_MEM(a,b,c,d)     nv_free_mem(&vm_list, a, b, c, d)
#  endif
#  undef VM_ALLOC_RECORD
#  define VM_ALLOC_RECORD(ptr, size, name) \
      if (ptr) { \
            vm_usage += size; \
            VM_PRINT("%s (0x%x: 0x%x): vm_usage is now 0x%x\n", \
                name, ptr, size, vm_usage); \
            VM_ADD_MEM(ptr, size, __FILE__, __LINE__); \
        }
#  undef VM_FREE_RECORD
#  define VM_FREE_RECORD(ptr, size, name) \
        if (ptr) { \
            vm_usage -= size; \
            VM_PRINT("%s (0x%x: 0x%x): vm_usage is now 0x%x\n", \
                name, ptr, size, vm_usage); \
            VM_FREE_MEM(ptr, size, __FILE__, __LINE__); \
        }
#endif

#if defined(KM_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  KM_PRINT
#    define KM_PRINT(str, args...)   printk(str, ##args)
#  endif
#  if defined(LIST_MEM_CHECK)
#    undef  KM_ADD_MEM
#    define KM_ADD_MEM(a,b,c,d)      nv_add_mem(&km_list, a, b, c, d)
#    undef  KM_FREE_MEM
#    define KM_FREE_MEM(a,b,c,d)     nv_free_mem(&km_list, a, b, c, d)
#  endif
#  undef KM_ALLOC_RECORD
#  define KM_ALLOC_RECORD(ptr, size, name) \
      if (ptr) { \
            km_usage += size; \
            KM_PRINT("%s (0x%x: 0x%x): km_usage is now 0x%x\n", \
                name, ptr, size, km_usage); \
            KM_ADD_MEM(ptr, size, __FILE__, __LINE__); \
        }
#  undef KM_FREE_RECORD
#  define KM_FREE_RECORD(ptr, size, name) \
        if (ptr) { \
            km_usage -= size; \
            KM_PRINT("%s (0x%x: 0x%x): km_usage is now 0x%x\n", \
                name, ptr, size, km_usage); \
            KM_FREE_MEM(ptr, size, __FILE__, __LINE__); \
        }
#endif

#if defined(FP_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  FP_PRINT
#    define FP_PRINT(str, args...)   printk(str, ##args)
#  endif
#if 0  // I had some problems tracking fp mem, so disable for now
#  if defined(LIST_MEM_CHECK)
#    undef  FP_ADD_MEM
#    define FP_ADD_MEM(a,b,c,d)      nv_add_mem(&fp_list, a, b, c, d)
#    undef  FP_FREE_MEM
#    define FP_FREE_MEM(a,b,c,d)     nv_free_mem(&fp_list, a, b, c, d)
#  endif
#endif
#endif

/* NV_VMALLOC has to allocate memory under the 4GB mark suitable for
 * DMA use with the hardware.  This has to be done differently for
 * different platforms and kernel versions.
 */
#if defined(NVCPU_IA64)
#define NV_VMALLOC(ptr, size) \
    { \
        (void *) (ptr) = vmalloc_dma(size); \
        VM_ALLOC_RECORD(ptr, size, "vm_alloc"); \
    }
#else
#define NV_VMALLOC(ptr, size) \
    { \
        (void *) (ptr) = vmalloc_32(size); \
        VM_ALLOC_RECORD(ptr, size, "vm_alloc"); \
    }
#endif

#define NV_VFREE(ptr, size) \
    { \
        VM_FREE_RECORD(ptr, size, "vm_alloc"); \
        vfree((void *) (ptr)); \
    }

#define NV_IOREMAP(ptr, physaddr, size) \
    { \
        (void *) (ptr) = ioremap(physaddr, size); \
        VM_ALLOC_RECORD(ptr, size, "vm_ioremap"); \
    }

#define NV_IOREMAP_NOCACHE(ptr, physaddr, size) \
    { \
        (void *) (ptr) = ioremap_nocache(physaddr, size); \
        VM_ALLOC_RECORD(ptr, size, "vm_ioremap_nocache"); \
    }

#define NV_IOUNMAP(ptr, size) \
    { \
        VM_FREE_RECORD(ptr, size, "vm_iounmap"); \
        iounmap(ptr); \
    }

/* only use this because GFP_KERNEL may sleep..
 * GFP_ATOMIC is ok, it won't sleep
 */
#define NV_KMALLOC(ptr, size) \
    { \
        (void *) (ptr) = kmalloc(size, GFP_KERNEL); \
        KM_ALLOC_RECORD(ptr, size, "km_alloc"); \
    }

#define NV_KMALLOC_ATOMIC(ptr, size) \
    { \
        (void *) (ptr) = kmalloc(size, GFP_ATOMIC); \
        KM_ALLOC_RECORD(ptr, size, "km_alloc_atomic"); \
    }  


#define NV_KFREE(ptr, size) \
    { \
        KM_FREE_RECORD(ptr, size, "km_free"); \
        kfree((void *) (ptr)); \
    }

#define NV_GET_FREE_PAGES(ptr, order) \
    { \
        (void *) (ptr) = __get_free_pages(NV_GFP_HW, order); \
    }
        
#define NV_FREE_PAGES(ptr, order) \
    { \
        free_pages(ptr, order); \
    }

#endif /* !defined NVWATCH */


#if defined (KERNEL_2_4)
#  define NV_IS_SUSER()                 suser()
#  define NV_PCI_DEVICE_NAME(dev)       ((dev)->name)
#  define NV_NUM_CPUS()                 smp_num_cpus
#  define NV_CLI()                      __cli()
#  define NV_SAVE_FLAGS(eflags)         __save_flags(eflags)
#  define NV_RESTORE_FLAGS(eflags)      __restore_flags(eflags)
#  define NV_MAY_SLEEP()                (!in_interrupt())
#  define NV_MODULE_PARAMETER(x)        MODULE_PARM(x, "i")
#endif

#if defined (KERNEL_2_6)
#  define NV_IS_SUSER()                 capable(CAP_SYS_ADMIN)
#  define NV_PCI_DEVICE_NAME(dev)       ((dev)->pretty_name)
#  define NV_NUM_CPUS()                 num_online_cpus()
#  define NV_CLI()                      local_irq_disable()
#  define NV_SAVE_FLAGS(eflags)         local_save_flags(eflags)
#  define NV_RESTORE_FLAGS(eflags)      local_irq_restore(eflags)
#  define NV_MAY_SLEEP()                (!in_interrupt() && !in_atomic())
#  define NV_MODULE_PARAMETER(x)        module_param(x, int, 0)
#endif

/* common defines */
#define GET_MODULE_SYMBOL(mod,sym)    (const void *) inter_module_get(sym)
#define PUT_MODULE_SYMBOL(sym)        inter_module_put((char *) sym)

#define NV_GET_PAGE_STRUCT(phys_page) virt_to_page(__va(phys_page))
#define NV_VMA_OFFSET(vma)            (((vma)->vm_pgoff) << PAGE_SHIFT)
#define NV_VMA_PRIVATE(vma)           ((vma)->vm_private_data)

#define NV_DEVICE_NUMBER(x)           minor((x)->i_rdev)
#define NV_IS_CONTROL_DEVICE(x)       (minor((x)->i_rdev) == 255)

#if defined(DEBUG)
#define NV_PRINT_AT(at)                                                    \
    nv_printf(NV_DBG_ERRORS, "nvidia lost alloc_t: 0x%08x  %4d  %2d  %05x 0x%08x 0x%08x 0x%08x\n", \
        (at->vma ? at->vma->vm_start : 0), at->num_pages, at->usage_count, \
        at->flags, at->page_table, at->key_mapping, at->agp_offset)
#else
#define NV_PRINT_AT(at)
#endif


#ifndef minor
# define minor(x) MINOR(x)
#endif

#ifndef IRQ_HANDLED
typedef void irqreturn_t;
#define IRQ_HANDLED
#endif

#if defined(KERNEL_2_6) && defined(AGPGART)
typedef struct agp_kern_info agp_kern_info;
typedef struct agp_memory agp_memory;
#endif

#if defined(CONFIG_DEVFS_FS)
#  if defined(KERNEL_2_6)
     typedef void* devfs_handle_t;
#    define NV_DEVFS_REGISTER(_name, _minor)                             \
     ({                                                                  \
         devfs_handle_t __handle = NULL;                                 \
         devfs_mk_cdev(MKDEV(NV_MAJOR_DEVICE_NUMBER, _minor),            \
                 S_IFCHR | S_IRUGO | S_IWUGO, _name);                    \
         __handle;                                                       \
     })

#    define NV_DEVFS_REMOVE_DEVICE(i) devfs_remove("nvidia%d", i)
#    define NV_DEVFS_REMOVE_CONTROL() devfs_remove("nvidiactl")
#  else // defined(KERNEL_2_4)
#    define NV_DEVFS_REGISTER(_name, _minor)                             \
     ({                                                                  \
         devfs_handle_t __handle =                                       \
             devfs_register(NULL, _name, DEVFS_FL_DEFAULT,               \
                     NV_MAJOR_DEVICE_NUMBER, _minor,                     \
                     S_IFCHR | S_IRUGO | S_IWUGO, &nv_fops, NULL);       \
         __handle;                                                       \
     })

#    define NV_DEVFS_REMOVE_DEVICE(i) devfs_unregister(nv_devfs_handles[i+1])
#    define NV_DEVFS_REMOVE_CONTROL() devfs_unregister(nv_devfs_handles[0])
#  endif // defined(KERNEL_2_4)
#endif // defined(CONFIG_DEVFS_FS)


/*
 * Linux 2.5 introduced the five argument version of remap_page_range, all
 * relevant releases to date use it. This version was backported to 2.4 
 * without means to identify the change, hence this hack.
 */
#if defined(KERNEL_2_6) || defined(REMAP_PAGE_RANGE_5)
#define NV_REMAP_PAGE_RANGE(a, b...)    remap_page_range(vma, a, ## b)
#elif defined(REMAP_PAGE_RANGE_4)
#define NV_REMAP_PAGE_RANGE(a, b...)    remap_page_range(a, ## b)
#else
#warning "conftest.sh failed, assuming old Linux 2.4 remap_page_range(4)!"
#define NV_REMAP_PAGE_RANGE(a, b...)    remap_page_range(a, ## b)
#endif

#if defined(pmd_offset_map)
#define NV_PMD_OFFSET(addres, pg_dir, pg_mid_dir) \
    { \
        pg_mid_dir = pmd_offset_map(pg_dir, address); \
    }
#define NV_PMD_UNMAP(pg_mid_dir) \
    { \
        pmd_unmap(pg_mid_dir); \
    }
#else
#define NV_PMD_OFFSET(addres, pg_dir, pg_mid_dir) \
    { \
        pg_mid_dir = pmd_offset(pg_dir, address); \
    }
#define NV_PMD_UNMAP(pg_mid_dir)
#endif

#define NV_PMD_PRESENT(pg_mid_dir) \
    ({ \
        if ( (pg_mid_dir) && (pmd_none(*pg_mid_dir))) { \
            NV_PMD_UNMAP(pg_mid_dir); pg_mid_dir = NULL; \
        } pg_mid_dir != NULL; \
    })

#if defined(pte_offset_atomic)
#define NV_PTE_OFFSET(addres, pg_mid_dir, pte) \
    { \
        pte = pte_offset_atomic(pg_mid_dir, address); \
        NV_PMD_UNMAP(pg_mid_dir); \
    }
#define NV_PTE_UNMAP(pte) \
    { \
        pte_kunmap(pte); \
    }
#elif defined(pte_offset)
#define NV_PTE_OFFSET(addres, pg_mid_dir, pte) \
    { \
        pte = pte_offset(pg_mid_dir, address); \
        NV_PMD_UNMAP(pg_mid_dir); \
    }
#define NV_PTE_UNMAP(pte)
#else
#define NV_PTE_OFFSET(addres, pg_mid_dir, pte) \
    { \
        pte = pte_offset_map(pg_mid_dir, address); \
        NV_PMD_UNMAP(pg_mid_dir); \
    }
#define NV_PTE_UNMAP(pte) \
    { \
        pte_unmap(pte); \
    }
#endif

#define NV_PTE_PRESENT(pte) \
    ({ \
        if (pte) { \
            if (!pte_present(*pte)) { \
                NV_PTE_UNMAP(pte); pte = NULL; \
            } \
        } pte != NULL; \
    })

#define NV_PTE_VALUE(pte) \
    ({ \
        unsigned long __pte_value = pte_val(*pte); \
        NV_PTE_UNMAP(pte); \
        __pte_value; \
    })

#define NV_PAGE_ALIGN(addr)             ( ((addr) + PAGE_SIZE - 1) / PAGE_SIZE)
#define NV_MASK_OFFSET(addr)            ( (addr) & (PAGE_SIZE - 1) )

#ifndef MAXMEM  /* temporary define for older kernels */
#define MAXMEM  (-PAGE_OFFSET - (64 * 1024 * 1024))
#endif

#if defined(NVCPU_IA64)
#define NV_GFP_HW (GFP_KERNEL | __GFP_DMA)
#define CACHE_FLUSH()
#else
#define NV_GFP_HW (GFP_KERNEL)
#define CACHE_FLUSH()  asm volatile("wbinvd":::"memory")
#endif

#if defined(NVCPU_X86) || defined(NVCPU_X86_64)
/* this isn't defined in some older kernel header files */
#define NV_CPU_INTERRUPT_FLAGS_BIT (1<<9)
#elif defined(NVCPU_IA64)
/* For whatever reason this is not defined an any header file I could
 * find.  From Intel IA64 Architecture Software Developers Manual Volume 2: 
 * IA64 System Architecture page 3-7 we have:
 */
#define NV_CPU_INTERRUPT_FLAGS_BIT (1<<14)
#else
#error define NV_CPU_INTERRUPT_FLAGS_BIT
#endif

static inline int NV_IRQL_IS_RAISED(void)
    {
        unsigned long int eflags;
        NV_SAVE_FLAGS(eflags);
        return !(eflags & NV_CPU_INTERRUPT_FLAGS_BIT);
    }
 
static inline int calc_order(int size)
    {
        int order = 0;
        while ( ((1 << order) * PAGE_SIZE) < (size))
        {
            order++;
        }
        return order;
    }

#if !defined (pgprot_noncached)
static inline pgprot_t pgprot_noncached(pgprot_t old_prot)
    {
        pgprot_t new_prot = old_prot;
        if (boot_cpu_data.x86 > 3)
            new_prot = __pgprot(pgprot_val(old_prot) | _PAGE_PCD);
        return new_prot;
    }
#endif

#if !defined(page_to_pfn)
#define page_to_pfn(page)  ((page) - mem_map)
#endif

/* On IA64 physical memory is partitioned into a cached and an
 * uncached view controlled by bit 63.  Set this bit when remapping
 * page ranges.  
 */
#if defined(NVCPU_IA64)
#define phys_to_uncached(addr) ((addr) | ((unsigned long) 1<<63))
#else
/* Some other scheme must be used on this platform */
#define phys_to_uncached(addr) (addr)
#endif

/*
 * An allocated bit of memory using NV_MEMORY_ALLOCATION_OFFSET
 *   looks like this in the driver
 */

typedef struct nv_pte_t {
    unsigned long   phys_addr;
    unsigned long   virt_addr;
    dma_addr_t      dma_addr;
} nv_pte_t;

typedef struct nv_alloc_s {
    struct nv_alloc_s *next;    
    struct vm_area_struct *vma;
    struct pci_dev *dev;
    unsigned int   usage_count;
    unsigned int   flags;
    unsigned int   agp_offset;
    unsigned int   num_pages;
    struct scatterlist *sg_list;        /* list of physical pages allocated */
    nv_pte_t      *page_table;          /* list of physical pages allocated */
    void          *key_mapping;         /* mapping used as a key for finding this nv_alloc_t */
                                        /*   may be the same as page_table                   */
    unsigned int   class;
    void          *priv_data;
} nv_alloc_t;

#define NV_ALLOC_TYPE_PCI      (1<<0)
#define NV_ALLOC_TYPE_AGP      (1<<1)
#define NV_ALLOC_TYPE_CONTIG   (1<<2)
#define NV_ALLOC_TYPE_KERNEL   (1<<3)
#define NV_ALLOC_TYPE_VMALLOC  (1<<4)

/* linux-specific version of old nv_state_t */
/* this is a general os-specific state structure. the first element *must* be
   the general state structure, for the generic unix-based code */
typedef struct {
    nv_state_t nv_state;

    struct pci_dev *dev;
    nv_alloc_t *alloc_queue;

    /* keep track of any pending bottom halfes */
    struct tasklet_struct tasklet;

    /* get a timer callback every second */
    struct timer_list rc_timer;

    /* per-device locking mechanism for access to core rm */
    spinlock_t rm_lock;
    int rm_lock_cpu;
    int rm_lock_count;

    spinlock_t heap_lock;

    /* lock for linux-specific data, not used by core rm */
    struct semaphore ldata_lock;

    /* lock for linux-specific alloc queue */
    struct semaphore at_lock;
} nv_linux_state_t;


/*
 * file-private data
 * hide a pointer to our data structures in a file-private ptr
 * there are times we need to grab this data back from the file
 * data structure..
 */

#define NV_EVENT_FIFO_SIZE 6

typedef struct
{
    void *nvptr;
    U032 num_events;
    U032 put, get;
    spinlock_t fp_lock;
    wait_queue_head_t waitqueue;
    nv_event_t *event_fifo;     // fifo for storing events
} nv_file_private_t;

#define FILE_PRIVATE(filep)     ((filep)->private_data)

#define NV_GET_NVFP(filep)      ((nv_file_private_t *) FILE_PRIVATE(filep))

/* for the card devices */
#define NVL_FROM_FILEP(filep) \
    ((nv_linux_state_t*) (NV_GET_NVFP(filep))->nvptr)

#define NV_GET_NVL_FROM_NV_STATE(nv) \
    ((nv_linux_state_t *) nv->os_state)

#define NV_STATE_PTR(nvl)   (&((nvl)->nv_state))

#define NV_ATOMIC_INC(data)  atomic_inc(&(data))
#define NV_ATOMIC_DEC(data)  atomic_dec(&(data))
#define NV_ATOMIC_READ(data) atomic_read(&(data))

/*
 * AMD Athlon processors expose a subtle bug in the Linux
 * kernel, that may lead to AGP memory corruption. Recent
 * kernel versions had a workaround for this problem, but
 * 2.4.20 is the first kernel to address it properly. The
 * page_attr API provides the means to solve the problem. 
 */
#if defined(NVCPU_X86) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20))
static inline void NV_SET_PAGE_ATTRIB_UNCACHED(nv_pte_t *page_ptr)
    {
        struct page *page = virt_to_page(page_ptr->virt_addr);
        change_page_attr(page, 1, PAGE_KERNEL_NOCACHE);
    }
static inline void NV_SET_PAGE_ATTRIB_CACHED(nv_pte_t *page_ptr)
    {
        struct page *page = virt_to_page(page_ptr->virt_addr);
        change_page_attr(page, 1, PAGE_KERNEL);
    }
#else
#define NV_SET_PAGE_ATTRIB_UNCACHED(page_list)
#define NV_SET_PAGE_ATTRIB_CACHED(page_list)
#endif

/*
 * Basic support for kgdb assertions.
 */
#if defined(CONFIG_X86_REMOTE_DEBUG)
#include <linux/gdb.h>

#define NV_ASSERT(message, condition) KGDB_ASSERT(message, condition)
#else
#define NV_ASSERT(message, condition)
#endif


#endif  /* _NV_LINUX_H_ */
